(* So, it looks like that in a suitable intuitionistic HOL (HA*ω can be one of those (and thus HAω, since it is a conservative extension)),
the proof EM <-> CB over PERs go through in the same way.
Let me know if you have ideas to simplify this proof :)
*)

Require Import Relations.
Require Import Arith.

Definition flip {A B C : Type} (R : A -> B -> C) := fun b a => R a b.

Arguments reflexive [_] _.
Arguments transitive [_] _.
Arguments symmetric [_] _.

Record PER :=
  {
     PER_carr : Type;
     PER_sim : relation PER_carr;
     PER_sym : symmetric PER_sim;
     PER_trans : transitive PER_sim;
  }.

Notation "x ≈ y" := (@PER_sim _ x y) (at level 54).
Notation "[ A ]" := (@PER_carr A).

Axiom magic : False.
Ltac magic := exfalso; exact magic.

Lemma PER_pr (A : PER) : forall a a' : [ A ], a ≈ a' -> a ≈ a.
  intros; apply PER_trans with a'; [| apply PER_sym]; assumption.
Qed.

Definition eq {A : Type} (a a' : A) := forall Z : A -> Prop, Z a -> Z a'.

Definition PER_discr (A : Type) : PER.
  exists A (fun x y => eq x y); magic.
Defined.

Definition PER_restr (A : PER) (P : [ A ] -> Prop) : PER.
  unshelve esplit.
  + exact [A].
  + exact (fun x y => P x /\ P y /\ x ≈ y). 
  + magic.
  + magic.
Defined.

Definition PER_codiscr (A : PER) : PER.
  unshelve esplit.
  + exact [A].
  + exact (fun x y => x ≈ x /\ y ≈ y).
  + magic.
  + magic.
Defined.

Definition ℕ : PER := PER_restr (PER_discr nat) (fun x => forall Z : nat -> Prop, Z 0 -> (forall y, Z y -> Z (S y)) -> Z x).

Lemma ℕ_zero_dom : (0 : [ℕ]) ≈ 0.
  repeat split; intro; auto.
Qed.

Lemma ℕ_succn_dom (n : [ℕ]) : n ≈ n -> (S n : [ℕ]) ≈ S n.
  intros [P [Q R]]; repeat split; [ | | intro; auto]; intros ? ? S; apply S, Q; assumption.
Qed.

Definition Bool : PER := PER_restr (PER_discr bool) (fun x => forall Z : bool -> Prop, Z true -> Z false -> Z x).

Lemma Bool_true_dom : (true : [Bool]) ≈ true.
  repeat split; intro; auto.
Qed.

Lemma Bool_false_dom : (false : [Bool]) ≈ false.
  repeat split; intro; auto.
Qed.

Definition PER_arr (A : PER) (B : PER) : PER.
Proof.
  exists ([A] -> [B]) (fun f g => forall a, a ≈ a -> f a ≈ g a); magic.
Defined.

Notation "A --> B" := (PER_arr A B) (at level 80, right associativity).

Definition PER_plus (A : PER) (B : PER) : PER.
Proof.
  exists ([A] + [B])%type (fun x y =>
                             match x, y with
                              inl a, inl a' => a ≈ a'
                            | inr b, inr b' => b ≈ b'
                            | _, _ => False
                             end /\
                            let i x : Prop:= forall Z : _ -> Prop,
                                      (forall a, a ≈ a -> Z (inl a)) -> (forall b, b ≈ b -> Z (inr b)) -> Z x
                            in i x /\ i y); magic.  
Defined.

Definition Ω : PER.
Proof.
  unshelve esplit; [exact Prop | exact (fun A B => A <-> B) | |]; firstorder.
Defined.

Definition functional {A B : PER} (R : [A --> B --> Ω]) :=
  forall a a' b b', a ≈ a' -> R a b -> R a' b' -> b ≈ b'.
Definition total {A B : PER} (R : [A --> B --> Ω]) := forall a, a ≈ a -> exists b, b ≈ b /\ R a b.

Definition PER_fun (A : PER) (B : PER) : PER := PER_restr (A --> B --> Ω) (fun f => functional f /\ total f).

Notation "A ---> B" := (PER_fun A B) (at level 80, right associativity).

Definition injective {A B : PER} (R : [A --> B --> Ω]) := functional (flip R).
Definition bitotal {A B : PER} (R : [A --> B --> Ω]) := total R /\ (total (flip R)).
Definition iso {A B : PER} (R : [A --> B --> Ω]) := functional R /\ injective R /\ bitotal R.

Definition bimpl (b b' : [Bool]) : [Bool] := if b' then if b then true else false else true.

Lemma bimpl_iff (b b' : [Bool]) : (b ≈ true -> b' ≈ true) <-> bimpl b' b ≈ true.

Definition Ninfty := PER_restr (ℕ ---> Bool) (fun z => forall n : [ℕ], n ≈ n -> z n false -> z (S n) false).

Notation "'ℕ∞'" := Ninfty.

Definition rel_of_fun {A B : Type} (f : A -> B) : A -> B -> Prop := fun a b => eq (f a) b.
Notation "! f" := (rel_of_fun f) (at level 3, left associativity).

Lemma rel_of_fun_dom {A B : PER} : (rel_of_fun : [(A --> B) --> A ---> B]) ≈ rel_of_fun.
Admitted.

Notation "'Z∞'" := (!(fun _ => false) : [ℕ∞]).
Notation "'S∞'" := ((fun p => (fun n => match n with
                                         0 => fun b => eq b true
                                       | S k => fun b => p k b
                                         end)) : [ℕ∞ --> ℕ∞]).
Notation "'pred∞'" := ((fun p n b => p (S n) b) : [ℕ∞ --> ℕ∞]).
Notation "'ω'" := (!(fun _ => true) : [ℕ∞]).

Reserved Notation "'ℕ∞_of_ℕ'".
Fixpoint Ninfty_of_N (n : [ℕ]) : [ℕ∞] :=
  match n with
   0 => Z∞
 | S k => S∞ (ℕ∞_of_ℕ k)
  end
where "'ℕ∞_of_ℕ'" := (Ninfty_of_N).


Lemma Ninfty_Ninfty_of_N : (ℕ∞_of_ℕ : [ℕ --> ℕ∞]) ≈ ℕ∞_of_ℕ.
Admitted.

Lemma Ninfty_zero_dom : Z∞ ≈ Z∞.
Admitted.

Lemma Ninfty_succ_dom : S∞ ≈ S∞.
Admitted.

Lemma Ninfty_pred_dom : pred∞ ≈ pred∞.
Admitted.

Lemma Ninfty_omega_dom : ω ≈ ω.
Admitted.

Lemma Bool_dec {b : [Bool]} : b ≈ b -> b ≈ true \/ b ≈ false.
  intros [H ?]; apply H; [left; apply Bool_true_dom|right; apply Bool_false_dom].
Qed.

Definition omniscient (A : PER) := forall f : [A ---> Bool], f ≈ f ->
  (exists a : [A], a ≈ a /\ f a true) \/ (forall a : [A], a ≈ a -> f a false). 

Definition ε_Ninfty (p : [ℕ∞ ---> Bool]) : [ℕ∞].
  intros ?.
  induction X.
  - intros ?.
    destruct X.
    + exact (p Z∞ false).
    + exact (p Z∞ true).
  - intros [|].
    + exact (IHX true /\ p (ℕ∞_of_ℕ (S X)) false).
    + exact (IHX false \/ p (ℕ∞_of_ℕ (S X)) true).
Defined.

Notation "'ε_ℕ∞'" := (ε_Ninfty : [(ℕ∞ ---> Bool) --> ℕ∞]).

Lemma Ninfty_ε_dom : ε_ℕ∞ ≈ ε_ℕ∞.
Admitted.

Lemma arr_dom {A B : PER} {f : [A --> B]} : f ≈ f -> forall {a : [A]}, a ≈ a -> f a ≈ f a.
Admitted.

Definition Bool_to_Prop : [Bool --> Ω] := fun b => b ≈ true.
Definition le (n m : [ℕ]) := Bool_to_Prop (n <=? m).
Definition lt (n m : [ℕ]) := Bool_to_Prop (n <? m).

Ltac reflexivity := intros; exact (fun _ x => x).
Ltac symmetry := symmetry || apply PER_sym.

Lemma ℕ_ind : forall {x : [ℕ]}, x ≈ x -> forall Z : [ℕ] -> Prop, Z 0 -> (forall k, k ≈ k -> Z k -> Z (S k)) -> Z x.
  intros.
  assert (Z x /\ x ≈ x) as G; [|destruct G; assumption].
  destruct H.
  apply H.
  - split; [assumption|apply ℕ_zero_dom].
  - intros ? [? ?]; split; [auto|apply ℕ_succn_dom; assumption].
Qed.

Lemma ℕ_case : forall {x : [ℕ]}, x ≈ x -> forall Z : [ℕ] -> Prop, Z 0 -> (forall k : [ℕ], k ≈ k -> Z (S k)) -> Z x.
  intros; apply ℕ_ind; auto.
Qed.

Lemma le_refl : forall n : [ℕ], n ≈ n -> le n n.
  intros ? e; apply (ℕ_ind e).
  + apply Bool_true_dom.
  + intros; unfold le in *; assumption.
Qed.

Lemma ε_Ninfty_nat_prod : forall (f : [ℕ∞ ---> Bool]) (n : [ℕ]), n ≈ n -> (forall m, le m n -> f (ℕ∞_of_ℕ m) false) -> ε_Ninfty f n true.
  intros f n Hn; apply (ℕ_ind Hn); intros.
  + specialize (H 0); simpl in *.
    apply H. apply le_refl, ℕ_zero_dom.
  + repeat split.
    - apply H0; intros. apply H1.
      magic.
    - apply H1, le_refl.
      apply ℕ_succn_dom, H.
Qed.

Definition conj' {A B : Prop} (a : A) (b : A -> B) : A /\ B := @conj _ _ a (b a).

Ltac asplit := apply conj'; [|intros ?].

Lemma restr_refl {A : PER} (P : [A] -> Prop) (a : [PER_restr A P]) : a ≈ a <-> (a : [A]) ≈ a /\ P a.
  split; [intros [? [_ ?]]| intros [? ?]; repeat split]; auto.
Qed.

Lemma restr_simpl {A : PER} {P : [A] -> Prop} (a b : [PER_restr A P]) : a ≈ a -> b ≈ b -> 
(a : [A]) ≈ (b : [A]) -> a ≈ b.
  intros.
  repeat asplit.
  + destruct H; assumption.
  + destruct H0; assumption.
  + apply H1.
Qed.

Lemma Bool_noconf : ((true : [Bool]) ≈ false) -> False.
  intros.
  destruct H as [? [? ?]].
  specialize (H1 (fun b => if b then True else False)).
  apply H1, I.
Qed.

Lemma le_ind' : forall {n} {k}, le n k -> forall (P : [ℕ] -> [ℕ] -> [Ω]), (forall n, n ≈ n -> P 0 n) -> (forall n k, n ≈ n -> k ≈ k -> le n k -> P n k -> P (S n) (S k)) -> n ≈ n -> k ≈ k -> P n k.
  intros ? ? ? ? ? ? H.
  revert k H0.
  apply (ℕ_ind H).
  + intros; apply H1; auto.
  + intros.
    revert H4; apply (ℕ_ind H5).
    - intros H'; apply PER_sym in H'.
      apply Bool_noconf in H'.
      destruct H'.
    - intros.
      apply H2; auto.
Qed.

Lemma le_ind : forall {k} {n} (p : le n k) (e : k ≈ k) (e' : n ≈ n) (P : [ℕ] -> [ℕ] -> [Ω]), (forall n, n ≈ n -> P n n) -> (forall n k, n ≈ n -> k ≈ k -> le n k -> P n k -> P n (S k)) -> P n k.
Admitted.

Lemma le_n_S_or : forall {k n : [ℕ]}, k ≈ k -> n ≈ n -> le n (S k) -> n ≈ S k \/ le n k.
  intros.
  apply (le_ind H1); auto.
Admitted.

Lemma Ninfty_of_N_true : forall k n : [ℕ], k ≈ k -> n ≈ n -> ℕ∞_of_ℕ n k (k <? n).
Admitted.

Lemma falseright : forall u : [ℕ∞], u ≈ u -> forall n, n ≈ n -> u n false -> forall k, k ≈ k -> le n k -> u k false.
  intros.
  revert H1.
  apply (le_ind H3 H2 H0); auto.
  intros.
  specialize (H6 H7).
  destruct H.
  apply H; auto.
Qed.

Lemma Ninfty_decr : forall (u : [ℕ∞]), u ≈ u -> forall (n m : [ℕ]), n ≈ n -> m ≈ m -> u n false -> le n m -> u m false.
  intros.
  revert n H0 H3 H2.
  apply (ℕ_ind H1).
  - intros.
    revert H2 H3; apply (ℕ_ind H0); auto.
    intros; exfalso.
    apply Bool_noconf, PER_sym, H5.
  - intros.
    destruct (le_n_S_or H0 H3); auto.
    + destruct H6 as [? [? ?]].
      apply H8; assumption.
    + destruct H.
      apply H; eauto.
Qed.

Lemma firstnonzeronat : forall (u : [ℕ∞]) (n : [ℕ]),  n ≈ n -> u ≈ u -> u n false -> (forall m, m ≈ m -> lt m n -> u m true) -> u ≈ ℕ∞_of_ℕ n.
  intros.
  apply restr_simpl; auto.
  + apply Ninfty_Ninfty_of_N; assumption.
  + apply restr_simpl; destruct H0 as [? [? ?]]; auto.
    - apply Ninfty_Ninfty_of_N; assumption.
    - {
        intros ? ? b ?.
        magic. (*TODO*)
      }
Qed.

Lemma ε_Ninfty_nat_min : forall (x : [ℕ]), x ≈ x -> forall (f : [ℕ∞ ---> Bool]), f ≈ f -> f (ℕ∞_of_ℕ x) true -> (forall y : [ℕ], y ≈ y -> lt y x -> f (ℕ∞_of_ℕ y) false) -> ε_ℕ∞ f ≈ ℕ∞_of_ℕ x.
  intros.
  apply firstnonzeronat; auto.
  + apply Ninfty_ε_dom; assumption.
  + simpl. revert H1; apply (ℕ_case H); simpl; auto.
  + intros ? Hm; apply (ℕ_ind Hm).
    - intro p; specialize (H2 _ ℕ_zero_dom p).
      apply H2.
    - intros; split.
      * apply H4.
        magic.
      * apply H2; auto.
        apply ℕ_succn_dom; assumption.
Qed.

Lemma ℕ_min_dec : forall (P : [ℕ --> Ω]) (x : [ℕ]), P x -> x ≈ x -> P ≈ P -> (forall n : [ℕ], n ≈ n -> P n \/ ~ P n) -> exists y, y ≈ y /\ P y /\ (forall z, z ≈ z -> lt z y -> ~ P z).
Admitted.

Lemma ε_Ninfty_no_nat : forall (f : [ℕ∞ ---> Bool]), f ≈ f ->
   (forall (x : [ℕ]), x ≈ x -> f (ℕ∞_of_ℕ x) false) -> ε_ℕ∞ f ≈ ω.
  intros ? ? ?; apply restr_simpl.
  + apply Ninfty_ε_dom, H.
  + magic. (*apply rel_of_fun_dom. eauto.*)
  + apply restr_simpl.
    - apply Ninfty_ε_dom, H.
    - magic.
    - intros ? ?.
      apply (ℕ_ind H1).
      * {
          split.
          + intros ? ? ?.
            simpl in H3.
            revert H3; destruct H2.
            apply H2; auto.
            intros?.
            exfalso; apply Bool_noconf; simpl in H5.
            destruct H as [[? _] _].
            specialize (H0 _ ℕ_zero_dom).
            eapply H; [|apply H5|apply H0]. apply restr_simpl.
            * magic.
            * apply Ninfty_Ninfty_of_N, ℕ_zero_dom.
            * apply restr_simpl.
              - magic.
              - apply Ninfty_Ninfty_of_N, ℕ_zero_dom.
              - intros ? ? ? ?.
                simpl; tauto.
          + intros.
            simpl.
            revert H3; destruct H2; apply H2; auto.
            * intros; apply (H0 _ ℕ_zero_dom).
            * intros e; exfalso; apply Bool_noconf,e,Bool_true_dom.
       }              
     * { intros; split; intro.
         + simpl in H5.
           revert H5; destruct H4; apply H4.
           - intros _; apply Bool_true_dom.
           - intros [?|?].
             * exfalso.
               apply Bool_noconf.
               magic. (* doable *)
             * exfalso. (*doable *)
               magic.
         + assert (a0 ≈ true).
           {
             apply H5, Bool_true_dom.
           }
           apply PER_sym in H6; destruct H6 as [_ [_ H6]]; apply H6.
           simpl; split.
           - apply H3; apply Bool_true_dom.
           - apply (H0 _ (ℕ_succn_dom _ H2)).
       }
Qed.

Lemma falsenat : forall u : [ℕ∞], u ≈ u -> forall n : [ℕ], n ≈ n -> u n false -> exists k : [ℕ], k ≈ k /\ u ≈ ℕ∞_of_ℕ k.
Admitted.

Lemma notnatω : forall u : [ℕ∞], u ≈ u -> (forall n : [ℕ], n ≈ n -> ~ u ≈ ℕ∞_of_ℕ n) -> u ≈ ω.
  intros.
  apply restr_simpl; [assumption| apply Ninfty_omega_dom |].
  apply restr_simpl; [destruct H as [? [? ?]]; assumption| apply Ninfty_omega_dom |].
  intros ? ? ? ?.
  split; intro.
  + destruct (Bool_dec H2).
    - apply PER_sym in H4; apply H4.
    - exfalso.
      specialize (falsenat _ H _ H1).
      intro; assert (u a false) by (destruct H4 as [_ [_ H4]]; apply H4,H3).
      specialize (H5 H6).
      destruct H5 as [? [? ?]]; eapply H0; eauto.
  + destruct (Bool_dec H2).
    - magic.
    - apply H4 in H3.
      exfalso; apply Bool_noconf, H3, Bool_true_dom.
Qed.  

Lemma Ninfty_ℕ_dense : forall p : [ℕ∞ ---> Bool], p ≈ p -> (forall x : [ℕ], x ≈ x -> p (ℕ∞_of_ℕ x) false) -> p ω false -> forall m : [ℕ∞], m ≈ m -> p m false.
  intros; assert (p m true \/ p m false) as D;[| destruct D].
  - destruct H as [[_ ?] _]. destruct (H _ H2) as [? [? ?]].
    revert H4; destruct H3; apply H3; tauto.
  - exfalso.
    assert (~ m ≈ ω).
    {
      intro.
      apply Bool_noconf; destruct H as [[? _] _].
      exact (H m ω _ _ H4 H3 H1).
    }
    assert (forall x : [ℕ], x ≈ x -> ~ m ≈ ℕ∞_of_ℕ x).
    {    
       intros ? ?.
       intros ?.       
       apply Bool_noconf.
       destruct H as [[? _] _].
       eapply H; [apply H6 | apply H3 | apply (H0 _ H5)].
    }
    apply H4.
    apply notnatω; assumption.
  - assumption.
Qed.

Lemma εfalse_N : forall f : [ℕ∞ ---> Bool], f ≈ f -> f (ε_ℕ∞ f) false -> forall n : [ℕ], n ≈ n -> f (ℕ∞_of_ℕ n) false.
      intros.
      assert (Hf' := H).
      destruct H as [[? ?] Hf].
      assert (ℕ∞_of_ℕ n ≈ ℕ∞_of_ℕ n) by (apply Ninfty_Ninfty_of_N, H1).
      destruct (H2 _ H3) as [? [? ?]].
      destruct (Bool_dec H4).
      * apply H6 in H5; clear x H4 H6; exfalso.
        {        
           destruct (ℕ_min_dec (fun z => f (ℕ∞_of_ℕ z) true) _ H5); auto.
           - destruct Hf as [[? ?] Hf]. intros ? ?. apply Hf.
             + apply Ninfty_Ninfty_of_N, H7.
             + apply Bool_true_dom.
           - intros; simpl.
             destruct (H2 _ (Ninfty_Ninfty_of_N _ H4)) as [? [? ?]].
             destruct (Bool_dec H6).
             + left; apply H8, H7.
             + right; intros ?; apply Bool_noconf.
               assert ((f : [ℕ∞ ---> Bool]) ≈ f) by (split; auto).
               eapply (H _ _ _ _ (Ninfty_Ninfty_of_N _ H4)); auto.
               apply H8, H7.
           - destruct H4 as [? [? ?]].
             specialize (ε_Ninfty_nat_min _ H4 _ Hf' H6).
             intros.
             assert (forall y : [ℕ], y ≈ y -> lt y x -> f (ℕ∞_of_ℕ y) false) by magic.
             specialize (H8 H9).
             apply Bool_noconf.
             apply PER_sym in H8.
             eapply (H _ _ _ _ H8); assumption.
        }
     * apply H6 in H5; assumption.
Qed.
  
Lemma Ninfty_omniscient : omniscient ℕ∞.
  intros ? ?.
  assert (total f) by magic.
  specialize (H0 _ (arr_dom Ninfty_ε_dom H)).
  destruct H0 as [? [? ?]].
  destruct (Bool_dec H0); [left|right].
  + exists (ε_Ninfty f).
    split.
    - apply Ninfty_ε_dom, H.
    - apply H2, H1.
  + apply H2 in H1; clear x H0 H2.
    intros; apply Ninfty_ℕ_dense; auto.
    - apply εfalse_N; assumption.
    - assert (H' := H); destruct H as [[fu ?] ?].
      assert (ε_Ninfty f ≈ ω).
      * apply ε_Ninfty_no_nat; try apply εfalse_N; auto.
      * destruct (H _ Ninfty_omega_dom) as [? [? ?]].
        specialize (fu _ _ _ _ H3 H1 H5).
        destruct fu as [_ [_ fu]].
        apply PER_sym in fu. apply fu, H5.
Qed.

Definition PER_sup (A : [Ω]) := PER_restr (PER_discr unit) (fun _ => A).

Definition section_retraction {A B : PER} (s : [A ---> B]) (r : [B ---> A]) :=
  forall a b a', a ≈ a -> b ≈ b -> a' ≈ a' -> s a b -> r b a' -> a ≈ a'.

Lemma plus_inl_dom {A B : PER} : (inl : [A --> PER_plus A B]) ≈ inl.
Admitted.

Lemma plus_inr_dom {A B : PER} : (inr : [B --> PER_plus A B]) ≈ inr.
Admitted.

Lemma crucial {A : [Ω]} {O : PER} : omniscient O -> forall s : [ PER_plus (PER_sup A) O ---> O ], s ≈ s -> forall r : [ O ---> PER_plus (PER_sup A) O ], r ≈ r -> section_retraction s r -> A \/ ~ A.
  intros.
  specialize (H (fun o b => forall v, r o v ->
                            match v with
                             inl _ => eq b true
                           | _ => eq b false
                            end)).
  destruct H.
  - magic.
  - destruct H as [a [ae ?]].
    left.
    destruct H1 as [[? ?] ?].
    specialize (H3 _ ae).
    destruct H3 as [? [? ?]].
    specialize (H _ H5).
    revert H; destruct H3 as [arf [H3 _]]; revert arf.
    apply H3; simpl.
    * intros _ [? _] _ _; assumption.
    * intros; exfalso; apply Bool_noconf.
      apply H6, Bool_true_dom.
  - right; intro.
    destruct H0 as [[? ?] ?].    
    destruct (H4 (inl tt)).
    + specialize (@plus_inl_dom (PER_sup A) O tt).
      intro p; apply p.
      repeat split; auto; intro; auto.
    + unfold section_retraction in H2.
      destruct H1 as [[? ?] ?], H6.
      specialize (H7 _ H6).
      destruct H7 as [? [? ?]].
      specialize (H _ H6 _ H10).
      assert ((inl  tt : [PER_plus (PER_sup A) O]) ≈ inl tt).
      {
        specialize (@plus_inl_dom (PER_sup A) O tt).
        intro p; apply p.
        repeat split; auto; intro; auto.
      }         
      specialize (H2 _ _ _ H11 H6 H7 H9 H10).
      destruct H2 as [? [_ ?]].
      revert H2 H; apply H12; auto.
      intros.
      apply Bool_noconf, H13, Bool_false_dom.
Qed.

Definition isomorphic A B := exists i : [A ---> B], iso i.

Definition CB := forall (A B : PER), forall (f : [A ---> B]) (g : [B ---> A]), f ≈ f -> g ≈ g -> injective f -> injective g -> isomorphic A B.

Theorem CB_PER : CB -> forall A, A \/ ~ A.
  intros.
  pose (f := (@inr [PER_sup A] [ℕ∞] : [ℕ∞ --> PER_plus (PER_sup A) ℕ∞])).
  pose (g (z : [PER_plus (PER_sup A) ℕ∞]) :=
             match z with
              inl _ => Z∞
            | inr m => S∞ m end).
  specialize (H _ _ !f !g).
  destruct H as [? ?].
  - apply rel_of_fun_dom, plus_inr_dom.
  - apply rel_of_fun_dom; subst g.
    intros ? ?.
    destruct H as [? [? _]].
    apply H0.
    + intros. apply Ninfty_zero_dom. 
    + intros; apply Ninfty_succ_dom; assumption.
  - intros ? ? ? ? ? ? ?.
    subst f.
    assert (H' := H).
    destruct H as [? [? _]].
    revert H H0.
    apply H2.
    + intros.
      compute in H3, H1.
      revert H0; apply H1.
      intros [].
    + intros.
      compute in H3, H1.
      revert H0; apply H1.
      apply (H3 (fun k => match k with inl _ => False | inr b'' => (b'' : [ℕ∞]) ≈ b' -> (b : [ℕ∞]) ≈ b' end)), (fun x => x).         
  - intros ? ? ? ? ? ? ?.
    subst g.
    assert (H' := H).
    destruct H as [? [? ?]].
    assert (H3' := H3).
    destruct H3' as [? [? H3']].
    specialize (H3' _ ℕ_zero_dom).
    specialize (PER_pr _ _ _ H3'); intros H3''.
    magic.
  - destruct H as [? [? [? ?]]].
    apply (fun k => crucial Ninfty_omniscient (flip x) k x).
    + magic.
    + magic.
    + intros ? ? ? ? ? ? ? ?.
      unfold functional in H.
      apply (H _ _ _ _ H4 H6 H7).
Qed.
