(* Let us do a version of CB:
   Assuming EM in Prop, A B : Type, f : A -> B -> Prop, g : B -> A -> Prop, injfun f ∧ injfun g
   then there exists a bitotal isomorphic relation between A and B.

   (the most general statement would be to get this over PERs)
   This proof holds in CC.
*)

Definition functional {A B : Type} (R : A -> B -> Prop) :=
  forall a b b', R a b -> R a b' -> b = b'.

Definition flip {A B C : Type} (R : A -> B -> C) := fun b a => R a b.
Definition injective {A B : Type} (R : A -> B -> Prop) := functional (flip R).
Definition total {A B : Type} (R : A -> B -> Prop) := forall a, exists b, R a b.
Definition bitotal {A B : Type} (R : A -> B -> Prop) := total R /\ (total (flip R)).
Definition iso {A B : Type} (R : A -> B -> Prop) := functional R /\ injective R /\ bitotal R.

Section CB.

Variable EM : forall A, A \/ ~ A.
Variable A B : Type.
Variable f : A -> B -> Prop.
Variable f_fun : functional f.
Variable f_inj : injective f.
Variable f_tot : total f.

Variable g : B -> A -> Prop.
Variable g_fun : functional g.
Variable g_inj : injective g.
Variable g_tot : total g.

(* Could be done in an impredicative way in CC *)
Inductive C : A -> Prop :=
 C_c : forall a b, g b a -> (forall a', f a' b -> C a') -> C a.

Definition i : A -> B -> Prop := fun a b => (C a /\ g b a) \/ (~ C a /\ f a b).

Lemma CBi : iso i.
  unfold i; repeat split.
  - intros ? ? ? [[? ?] | [? ?]] [[? ?] | [? ?]];
    try ((exfalso; auto; fail) || eauto using g_inj, f_fun).
  - intros ? ? ?  [[? ?] | [? ?]] [[? ?] | [? ?]]; eauto using g_fun, g_inj, f_inj.
    + destruct H. exfalso; congruence.
      exfalso; apply H1, H3.
      eapply g_inj in H.
      rewrite H; [apply H2|]; assumption.
    + exfalso; apply H.
      destruct H1. apply H3.
      assert (b0 = a) by (eapply g_inj; eassumption); subst; assumption.
    + eapply f_inj; eassumption.
  - intro a; destruct (EM (C a)).
    + destruct H.
      exists b; left; split; [econstructor|]; eassumption.
    + destruct (f_tot a) as [b Hb]; exists b; eauto.
  - intro b.
    destruct (EM (exists a, ~ C a /\ f a b)).
    + destruct H as [a [? ?]]; exists a; right; eauto.
    + destruct (g_tot b) as [a Ha]; exists a; left; split; eauto.
      apply (C_c _  _ Ha).
      intros.
      destruct (EM (C a')); auto.
      exfalso; apply H; eauto.
Qed.

Theorem CB : exists j : A -> B -> Prop, iso j.
  exists i; apply CBi.
Qed.

End CB.

Check CB.
